/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Script
Version:   1.0
Status:    Stable
ID:        ID_SCRIPT
Category:  Data
Date:      December 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  script, dml, file, execute, run, scripting, object, class, processor, compile
Short:     The Script class manages the processing of object script files.

###DESCRIPTION###
<p>The Script class is provided for the purpose of processing object scripts.
This class specifically supports the Dynamic Markup Language (DML), but
other script types can be supported through sub-classes.  DML is the most
common language used throughout the Athene system and the files can be
identified by a &quot;.dml&quot; file extension.  For a complete introduction
on the DML scripting standard, you will need to refer to the
<a href="http://www.rocklyte.com/dml_whitepaper.html">DML Whitepaper</a> for
all necessary details.</p>

<p>To run an object script, you need to set the
<a href="#Location">Location</a> field and then Activate the Script object.
If a particular object script supports arguments, you can set values for them
via the SetUnlistedField action.  Activating an object script will often
cause a large amount of objects to be generated.  If you need to find any
of these objects, read the <a href="#ObjectList">ObjectList</a> field to
retrieve a list of all objects that were created by the activation
process.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/files/file.h>
#include <pandora/data/config.h>
#include <pandora/data/script.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>

#include "script.h"

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_SCRIPT 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_SCRIPT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "December 2003",
   "Dynamic Markup Language"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR ScriptClass = 0;
static OBJECTPTR StringsMod = 0;
static OBJECTPTR FileMod = 0;
static struct Script *glScript = 0, *glPrevScript = 0;
static struct XMLTag *glChildTags = 0;
static STRING glRepeatCount = 0;
static UBYTE glLocale[4] = "eng";

/*****************************************************************************
** Script class.
*/

static ERROR GET_Cache(struct Script *, LONG *);
static ERROR GET_LocalArgs(struct Script *, MEMORYID **);
static ERROR GET_Location(struct Script *, STRING *);
static ERROR GET_ObjectList(struct Script *, OBJECTID **);
static ERROR GET_Procedure(struct Script *, STRING *);
static ERROR GET_Service(struct Script *, STRING *);
static ERROR GET_String(struct Script *, STRING *);

static ERROR SET_Cache(struct Script *, LONG *);
static ERROR SET_Location(struct Script *, STRING);
static ERROR SET_Name(struct Script *, STRING);
static ERROR SET_Procedure(struct Script *, STRING);
static ERROR SET_RegisterObject(struct Script *, OBJECTID *);
static ERROR SET_Service(struct Script *, STRING);
static ERROR SET_Static(struct Script *, LONG *);
static ERROR SET_String(struct Script *, STRING);
static ERROR SET_Target(struct Script *, OBJECTID *);

static struct FieldDef ScriptFlags[] = {
   { "OBJECTLIST", SCF_OBJECTLIST }, { "PROGRAM", SCF_PROGRAM },
   { "BATCH",      SCF_BATCH },      { "PERSISTENT", SCF_PERSISTENT },
   { "CACHE",      SCF_CACHE },
   { NULL, NULL }
};

static struct FieldArray ScriptFields[] = {
   { "Target",         0, FDF_OBJECTID|FDF_RW,  0, NULL, SET_Target },
   { "PrimaryObject",  0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "Output",         0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "Static",         0, FDF_LONG|FDF_INIT,    0, NULL, SET_Static },
   { "TotalArgs",      0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "TotalObjects",   0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Repeat",         0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "RepeatStart",    0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RI, (LONG)&ScriptFlags, NULL, NULL },
   { "Error",          0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "FileSize",       0, FDF_LARGE|FDF_R,      0, NULL, NULL },
   { "TimeStamp",      0, FDF_LARGE|FDF_R,      0, NULL, NULL },
   /*** Virtual Fields ***/
   { "Cache",          0, FDF_LONG|FDF_RW,    0, GET_Cache,      SET_Cache },
   { "Location",       0, FDF_STRING|FDF_RI,  0, GET_Location,   SET_Location },
   { "Procedure",      0, FDF_STRING|FDF_RW,  0, GET_Procedure,  SET_Procedure },
   { "Service",        0, FDF_STRING|FDF_RW,  0, GET_Service,    SET_Service },
   { "LocalArgs",      0, FDF_POINTER|FDF_R,  0, GET_LocalArgs,  NULL },
   { "Name",           0, FDF_STRING|FDF_RW,  0, NULL,           SET_Name },
   { "ObjectList",     0, FDF_POINTER|FDF_R,  0, GET_ObjectList, NULL },
   { "RegisterObject", 0, FDF_OBJECTID|FDF_W, 0, NULL,           SET_RegisterObject },
   { "Src",            0, FDF_STRING|FDF_RI,  0, GET_Location,   SET_Location },
   { "Statement",      0, FDF_STRING|FDF_RW,  0, GET_String,     SET_String },
   { "String",         0, FDF_STRING|FDF_RW,  0, GET_String,     SET_String },
   END_FIELD
};

/*****************************************************************************
** Script actions.
*/

static ERROR SCRIPT_Activate(struct Script *, APTR);
static ERROR SCRIPT_AccessObject(struct Script *, APTR);
static ERROR SCRIPT_ClosingTag(struct Script *, APTR);
static ERROR SCRIPT_Free(struct Script *, APTR);
static ERROR SCRIPT_GetUnlistedField(struct Script *, struct acGetUnlistedField *);
static ERROR SCRIPT_Init(struct Script *, APTR);
static ERROR SCRIPT_NewChild(struct Script *, struct acNewChild *);
static ERROR SCRIPT_NewObject(struct Script *, APTR);
static ERROR SCRIPT_OpeningTag(struct Script *, APTR);
static ERROR SCRIPT_ReleaseObject(struct Script *, APTR);
static ERROR SCRIPT_SetUnlistedField(struct Script *, struct acSetUnlistedField *);

static struct ActionArray ScriptActions[] = {
   { AC_Activate,         SCRIPT_Activate },
   { AC_AccessObject,     SCRIPT_AccessObject },
   { AC_ClosingTag,       SCRIPT_ClosingTag },
   { AC_Free,             SCRIPT_Free },
   { AC_GetUnlistedField, SCRIPT_GetUnlistedField },
   { AC_Init,             SCRIPT_Init },
   { AC_NewChild,         SCRIPT_NewChild },
   { AC_NewObject,        SCRIPT_NewObject },
   { AC_OpeningTag,       SCRIPT_OpeningTag },
   { AC_ReleaseObject,    SCRIPT_ReleaseObject },
   { AC_SetUnlistedField, SCRIPT_SetUnlistedField },
   { NULL, NULL }
};

/*****************************************************************************
** If statement data.
*/

#define COND_NOT_EQUAL     1
#define COND_EQUAL         2
#define COND_LESS_THAN     3
#define COND_LESS_EQUAL    4
#define COND_GREATER_THAN  5
#define COND_GREATER_EQUAL 6

static struct FieldDef ConditionTable[] = {
   { "<>",               COND_NOT_EQUAL },
   { "!=",               COND_NOT_EQUAL },
   { "NOT",              COND_NOT_EQUAL },
   { "=",                COND_EQUAL },
   { "==",               COND_EQUAL },
   { "IS",               COND_EQUAL },
   { "<",                COND_LESS_THAN },
   { "LESS",             COND_LESS_THAN },
   { "LESS THAN",        COND_LESS_THAN },
   { "<=",               COND_LESS_EQUAL },
   { "LESS OR EQUAL",    COND_LESS_EQUAL },
   { ">",                COND_GREATER_THAN },
   { "GREATER",          COND_GREATER_THAN },
   { "GREATER THAN",     COND_GREATER_THAN },
   { ">=",               COND_GREATER_EQUAL },
   { "GREATER OR EQUAL", COND_GREATER_EQUAL },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes.
*/

static LONG ExtractField(STRING, STRING);
static void ExtractValue(STRING, STRING, LONG);
static LONG LoadScript(struct Script *);
static LONG NextElement(STRING str);
static struct XMLTag * NextTag(struct Script *, struct XMLTag *, OBJECTID, STRING, STRING, STRING);
static OBJECTID ProcessObject(struct Script *, OBJECTID, STRING, STRING, WORD);
static void SetDMLField(OBJECTPTR, STRING, STRING);
static LONG TestStatement(STRING, STRING, LONG);
static void TranslateArgs(struct Script *, STRING, STRING, LONG);
static void WriteField(struct Script *, OBJECTPTR, STRING, STRING);

#define KWD_IGNORED   1
#define KWD_PROCESSED 2

/****************************************************************************/

static BYTE LCase(BYTE Case)
{
  if ((Case >= 'A') AND (Case <= 'Z')) return(Case - 'A' + 'a');
  else return(Case);
}

static STRING AccessArg(struct Script *Self, LONG Index)
{
   STRING arg;

   if (Self->Head.MemFlags & MEM_PUBLIC) {
      if (AccessMemory(Self->LocalArgs[Index], MEM_READ, &arg) IS ERR_Okay) {
         return(arg);
      }
      else return(NULL);
   }
   else return((STRING)Self->LocalArgs[Index]);
}

static void ReleaseArg(struct Script *Self, STRING Arg)
{
   if (Self->Head.MemFlags & MEM_PUBLIC) ReleaseMemory(Arg);
}

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR config;

   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Get the user's locale for translation purposes ***/

   if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
      FID_Location|TSTRING, "user:config/locale.cfg",
      TAGEND) IS ERR_Okay) {
      struct mtReadConfig read;
      read.Section = "LOCALE";
      read.Item    = "Language";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         if (StrLength(read.Result) IS 3) {
            StrCopy(read.Result, glLocale, sizeof(glLocale));
            StrLower(glLocale);
         }
      }
      Action(AC_Free, config, NULL);
   }

   /*** Create the Script class ***/

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ScriptClass, NULL,
      FID_BaseClassID|TLONG,       ID_SCRIPT,
      FID_Version|TFLOAT,          VER_SCRIPT,
      FID_Name|TSTRING,            "Script",
      FID_Category|TLONG,          CCF_DATA,
      FID_FileExtension|TSTRING,   "dml;xml",
      FID_FileDescription|TSTRING, "Dynamic Markup Language",
      FID_Actions|TPTR,            ScriptActions,
      FID_Fields|TPTR,             ScriptFields,
      FID_Size|TLONG,              sizeof(struct Script),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ScriptClass) { Action(AC_Free, ScriptClass, NULL); ScriptClass = NULL; }
   if (StringsMod)  { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (FileMod)     { Action(AC_Free, FileMod , NULL); FileMod = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Script: AccessObject
*/

static ERROR SCRIPT_AccessObject(struct Script *Self, APTR Void)
{
   if (Self->LocationMID) {
      if (AccessMemory(Self->LocationMID, MEM_READWRITE, (void **)&Self->Location) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   if (Self->ProcedureMID) {
      if (AccessMemory(Self->ProcedureMID, MEM_READWRITE, (void **)&Self->Procedure) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   if (Self->LocalArgsMID) {
      if (AccessMemory(Self->LocalArgsMID, MEM_READWRITE, (void **)&Self->LocalArgs) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   if (Self->StringMID) {
      if (AccessMemory(Self->StringMID, MEM_READWRITE, (void **)&Self->String) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessMemory));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Processes a Script.
###END###

*****************************************************************************/

static ERROR SCRIPT_Activate(struct Script *Self, APTR Void)
{
   #undef SIZE_GLOBAL
   #define SIZE_GLOBAL 4096
   struct XMLTag **xmllist, *xmltag, *dmltag, *repeattag;
   STRING Translation, TagBuffer, arg;
   BYTE RepeatStr[30];
   OBJECTID OwnerID;
   OBJECTPTR owner;
   LONG i, j, pos;
   ERROR error;

   DPrintF("~Activate()","[Script:%d] %s, Target: %d", GetUniqueID(Self), Self->Location, Self->TargetID);

   Self->Error = ERR_Okay;
   Self->ScriptOwnerID = Self->Head.OwnerID; /* Necessary becuase we will be using SetOwner() later */
   Self->prvQuit = FALSE;

   Translation = NULL;
   TagBuffer   = NULL;
   if ((error = AllocMemory(SIZE_GLOBAL, MEM_STRING|MEM_NOCLEAR, (APTR *)&Translation, NULL)) != ERR_Okay) goto exit;
   if ((error = AllocMemory(SIZE_GLOBAL, MEM_STRING|MEM_NOCLEAR, (APTR *)&TagBuffer, NULL)) != ERR_Okay) goto exit;

   if ((GetField(Self, FID_Service, FT_STRING, &Self->Service) IS ERR_Okay) AND (Self->Service)) {
      struct msgCallService *servicemsg;
      struct mtCallService call;
      OBJECTID ServiceID;
      UBYTE msgbuffer[(sizeof(struct Message) + sizeof(struct msgCallService))];
      UBYTE buffer[1000];

      DPrintF("Activate:","Sending the script to service \"%s\".", Self->Service);

      if (FastFindObject(Self->Service, ID_SERVICE, &ServiceID, 1, NULL) IS ERR_Okay) {
         buffer[0] = '#';
         IntToStr(Self->Head.TaskID, buffer+1, sizeof(buffer)-2);
         SetUnlistedField(Self, "caller", buffer);

         if (Self->TotalArgs > 0) {
            ((LONG *)buffer)[0] = Self->TotalArgs;
            pos = 4;
            for (i=0; i < Self->TotalArgs; i++) {
               if ((arg = AccessArg(Self, i))) {
                  for (j=1; arg[j] != '}'; j++) buffer[pos++] = arg[j];
                  buffer[pos++] = '=';
                  j += 2;
                  while (arg[j]) buffer[pos++] = arg[j++];
                  buffer[pos++] = 0;
                  ReleaseArg(Self, arg);
               }
            }
            call.Args    = buffer;
            call.ArgSize = pos;
         }
         else {
            call.Args    = NULL;
            call.ArgSize = NULL;
         }
         call.MessageQueue = GetResource(RES_MESSAGEQUEUE);
         if (ActionMsg(MT_CallService, ServiceID, &call) IS ERR_Okay) {

            /*** Wait for the service to respond within the timeout period (5 second default) ***/

            SetResource(RES_TIMEOUT, 5000);
            if (GetMessage(NULL, MSGID_SERVICE, MF_WAIT|MF_TIMEOUT, &msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
               servicemsg = (struct msgCallService *)(msgbuffer + sizeof(struct Message));

               /*** The service has responded, so it attempted to execute the script ***/

               if (servicemsg->Error IS ERR_Okay) {
                  if ((Self->PrimaryObjectID = servicemsg->ObjectID)) {
                     DPrintF("Activate:","[Script:%d] Object #%d was specified as the primary object.", Self->Head.UniqueID, Self->PrimaryObjectID);
                     SetField(Self, FID_RegisterObject, FT_LONG, &Self->PrimaryObjectID);

                     if (Self->ChildFeedback) {
                        DPrintF("Activate:","[Script] Proceeding with child feedback processing.");

                        Self->ChildFeedback(Self->PrimaryObjectID, glChildTags, TagBuffer, Translation);
                        Self->ChildFeedback = NULL;

                     }
                  }
                  else DPrintF("@Activate:","[Script:%d] Warning - the script did not specify a primary object.", Self->Head.UniqueID);

                  error = ERR_Okay;
                  goto exit;
               }
               else {
                  DPrintF("@Activate:","[Script:%d] The service returned an execution error.", Self->Head.UniqueID);
                  error = ObjectError(Self, ERH_Activate, servicemsg->Error);
                  goto exit;
               }
            }
            else DPrintF("@Activate:","[Script:%d] The service did not respond to the execution request.", Self->Head.UniqueID);

            /* The service has not responded in the allotted time, so assume
            ** failure and continue normal execution to launch the script in
            ** our own task space.
            */
         }
         else DPrintF("@Activate:","[Script:%d] Failed to send an execution message to the service.", Self->Head.UniqueID);
      }
      else DPrintF("@Activate:","[Script:%d] Service \"%s\" does not exist.", Self->Head.UniqueID, Self->Service);
   }

   error = ERR_Okay;

   Self->GeneratedID = AllocateID(IDTYPE_GLOBAL);

   /* For safety reasons, set the script owner to the current Task
   ** prior to the execution process.  Once complete, we will change
   ** back to the original owner.
   */

   OwnerID = GetOwner(Self);
   SetOwner(Self, CurrentTask());

   /*** Reset the primary object ***/

   Self->PrimaryObjectID = NULL;

   /*** If necessary, reload the script from scratch ***/

   if ((Self->Static) OR (!Self->XML)) {
      if ((Self->XML) AND (Self->Flags & SCF_CACHE)); /* Do not reload if cache-mode is enabled */
      else if (LoadScript(Self) != ERR_Okay) {
         error = ObjectError(Self, ERH_Activate, ERR_Failed);
         goto exit;
      }
   }

   if ((GetFields(Self->XML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) AND (xmllist)) {

      /* Search for a dml tag and make sure that it matches our procedure if
      ** we have specified one.
      */

      dmltag = NULL;
      for (i=0; xmllist[i]; i++) {
         if (StrCompare("dml", xmllist[i]->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (Self->Procedure) {
               for (j=1; j < xmllist[i]->TotalAttributes; j++) {
                  if (StrCompare("procedure", xmllist[i]->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     if (StrCompare(Self->Procedure, xmllist[i]->Attributes[j].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        dmltag = xmllist[i];
                        break;
                     }
                  }
               }
            }
            else dmltag = xmllist[i];

            if (dmltag) break;
         }
      }

      if (dmltag) xmltag = dmltag->Child;
      else xmltag = NULL;

      if (dmltag) {
         /*** Process arguments that might be in the <dml> tag ***/

         for (j=1; j < dmltag->TotalAttributes; j++) {
             if (StrCompare("translate-content", dmltag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                DPrintF("Activate:","[Script] Content translation is on.");
                Self->prvTranslateAll = TRUE;
             }
             else if (StrCompare("exitonerror", dmltag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                DPrintF("Activate:","[Script] Exit-On-Error is on.");
                Self->Flags |= SCF_EXITONERROR;
             }
             else if (StrCompare("persistent", dmltag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                DPrintF("Activate:","[Script] Script is persistent.");
                Self->Flags |= SCF_PERSISTENT;
             }
         }
      }
      else {
         /* If a procedure was specified but wasn't found,
         ** abort the activation routine.
         */

         if (Self->Procedure) {
            DPrintF("@Activate:","[Script] Failed to find procedure \"%s\".", Self->Procedure);
            error = ERR_Search;
            goto exit;
         }

         /* If the embedded flag was specified and no <dml> tag was found, abort the
         ** routine.  Otherwise, just print a warning to the debugger.
         */

         if (Self->Flags & SCF_EMBEDDED) {
            error = ObjectError(Self, ERH_Activate, ERR_BadData);
            goto exit;
         }

         xmltag = xmllist[0];
      }

      Self->LoopIndex = 0;
      Self->LoopEnd = 0;

      if (Self->Repeat > 0) {
         DPrintF("Script:","I will execute the DML file %d times.", Self->Repeat - Self->RepeatStart);

         for (i=Self->RepeatStart; i < Self->Repeat; i++) {
            IntToStr(i, RepeatStr, 30);
            repeattag = xmltag;
            while (repeattag) repeattag = NextTag(Self, repeattag, Self->TargetID, RepeatStr, TagBuffer, Translation);
         }
      }
      else {
         while (xmltag) {
            xmltag = NextTag(Self, xmltag, Self->TargetID, NULL, TagBuffer, Translation);
         }
      }

      Self->ActivationCount++;

      /* Change back to the original owner if it still exists.  If it
      ** doesn't, commit suicide.
      */

      if (OwnerID) {
         if (AccessObject(OwnerID, 5000, &owner) IS ERR_Okay) {
            SetOwner(Self, owner);
            ReleaseObject(owner);
         }
         else {
            DPrintF("Activate:","[Script:%d] Owner #%d no longer exists - committing suicide.", Self->Head.UniqueID, OwnerID);
            Action(AC_Free, Self, NULL);
         }
      }
   }
   else {
      DPrintF("@Activate:","[Script] Failed to read XML Tags from XML object $%x.", Self->XML);
      error = ERR_GetField;
   }

exit:
   if (error IS ERR_Okay) error = Self->Error;

   if (TagBuffer)   FreeMemory(TagBuffer);
   if (Translation) FreeMemory(Translation);
   StepBack();
   return(error);
}

/****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If a Script object is non-static, calling this action will cause it to self-destruct.
###END###

*****************************************************************************/

static ERROR SCRIPT_ClosingTag(struct Script *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      DPrintF("~ClosingTag()","[Script:%d] Non-static script, self-destructing...", Self->Head.UniqueID);
      Action(AC_Free, Self, NULL);
      StepBack();
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Script: Free
*/

static ERROR SCRIPT_Free(struct Script *Self, APTR Void)
{
   LONG i;

   if (Self->LocalArgsMID) {
      if (!Self->LocalArgs) GET_LocalArgs(Self, &Self->LocalArgs);

      if (Self->LocalArgs) {
         for (i=0; i < Self->TotalArgs; i++) {
            if (Self->LocalArgs[i]) {
               if (Self->Head.MemFlags & MEM_PUBLIC) FreeMemoryID(Self->LocalArgs[i]);
               else FreeMemory((APTR)Self->LocalArgs[i]);
            }
         }
         ReleaseMemory(Self->LocalArgs);
         Self->LocalArgs = NULL;
      }

      FreeMemoryID(Self->LocalArgsMID);
      Self->LocalArgsMID = NULL;
   }

   if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

   if (Self->ObjectList)    { ReleaseMemory(Self->ObjectList);   Self->ObjectList = NULL; }
   if (Self->ObjectListMID) { FreeMemoryID(Self->ObjectListMID); Self->ObjectListMID = NULL; }

   if (Self->Procedure)    { ReleaseMemory(Self->Procedure);   Self->Procedure = NULL; }
   if (Self->ProcedureMID) { FreeMemoryID(Self->ProcedureMID); Self->ProcedureMID = NULL; }

   if (Self->Service)    { ReleaseMemory(Self->Service);   Self->Service = NULL; }
   if (Self->ServiceMID) { FreeMemoryID(Self->ServiceMID); Self->ServiceMID = NULL; }

   if (Self->String)    { ReleaseMemory(Self->String);   Self->String = NULL; }
   if (Self->StringMID) { FreeMemoryID(Self->StringMID); Self->StringMID = NULL; }

   /*** Free the XML object ***/

   if (Self->XML) {
      Action(AC_Free, Self->XML, NULL);
      Self->XML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Script arguments can be retrieved through this action.
###END###

*****************************************************************************/

static ERROR SCRIPT_GetUnlistedField(struct Script *Self, struct acGetUnlistedField *Args)
{
   STRING arg;
   LONG i, j, k;
   UBYTE a;

   for (i=0; i < Self->TotalArgs; i++) {
      if ((arg = AccessArg(Self, i))) {
         for (j=0; Args->Field[j]; j++) {
            if ((Args->Field[j] >= 'A') AND (Args->Field[j] <= 'Z')) a = Args->Field[j] - 'A' + 'a';
            else a = Args->Field[j];

            if ((arg[j+1] >= 'A') AND (arg[j+1] <= 'Z')) {
               if (a != arg[j+1] - 'A' + 'a') break;
            }
            else if (a != arg[j+1]) break;
         }

         if (arg[j+1] IS '}') {
            k = 0;
            j += 3;
            while ((arg[j]) AND (k < Args->Size-1)) Args->Buffer[k++] = arg[j++];
            Args->Buffer[k] = 0;
            ReleaseArg(Self, arg);
            return(ERR_Okay);
         }
         ReleaseArg(Self, arg);
      }
   }

   return(ERR_UnsupportedField);
}

/*****************************************************************************
** Script: Init
*/

static ERROR SCRIPT_Init(struct Script *Self, APTR Void)
{
   BYTE buffer[60];
   ERROR error;

   if ((!Self->Location) AND (!Self->String)) return(ObjectError(Self, ERH_Init, ERR_MissingLocation));

   /*** Define the target if it has not been set already ***/

   if (!Self->TargetID) {
      DPrintF("Init:","[Script] Target not set, defaulting to owner #%d.", Self->Head.OwnerID);
      Self->TargetID = Self->Head.OwnerID;
   }

   /*** Set the owner argument if it has not been set already ***/

   if (GetUnlistedField(Self, "owner", buffer, sizeof(buffer)) != ERR_Okay) {
      buffer[0] = '#';
      IntToStr(Self->TargetID, buffer+1, sizeof(buffer)-2);
      SetUnlistedField(Self, "owner", buffer);
   }

   /*** Set the output argument if it has not been set already ***/

   if (GetUnlistedField(Self, "output", buffer, sizeof(buffer)) != ERR_Okay) {
      buffer[0] = '#';
      IntToStr(Self->OutputID, buffer+1, sizeof(buffer)-2);
      SetUnlistedField(Self, "output", buffer);
   }

   /* Check the format of the file data to see if it is XML based.  If
   ** we cannot interpret the data, return ERR_NoSupport so that other
   ** script classes can use it.
   */







   /* Non-static scripts are always loaded during initialisation.
   ** Static scripts are usually re-loaded for each activation,
   ** so we do not deal with them here.
   */

   if (Self->Static IS FALSE) {
      if ((error = LoadScript(Self)) IS ERR_Okay) {
         DPrintF("2Script:","DML script initialised successfully.");
         return(ERR_Okay);
      }
      else {
         SCRIPT_Free(Self, NULL); /* Free allocations only - not the object itself */
         return(ObjectError(Self, ERH_Init, error));
      }
   }
   else return(ERR_Okay); /* Static script - return immediately */
}

/*****************************************************************************
** Script: NewChild
*/

static ERROR SCRIPT_NewChild(struct Script *Self, struct acNewChild *Args)
{
   OBJECTID OwnerID;
   OBJECTPTR Owner, NewChild;

   if ((Self->PrimaryObjectID) AND (Args->NewChildID < 0)) {
      OwnerID = Self->PrimaryObjectID;
   }
   else return(ERR_Okay);

   if (AccessObject(OwnerID, 2000, &Owner) IS ERR_Okay) {
      if (AccessObject(Args->NewChildID, 2000, &NewChild) IS ERR_Okay) {
         DPrintF("NewChild()","[Script] Child %d will be passed to Owner %d.", Args->NewChildID, OwnerID);
         SetOwner(NewChild, Owner);
         ReleaseObject(NewChild);
         ReleaseObject(Owner);
         return(ERR_OwnerPassThrough);
      }
      else {
         ReleaseObject(Owner);
         DPrintF("@NewChild:","[Script] Could not gain access to new child object %d.", Args->NewChildID);
         return(ERR_ExclusiveDenied);
      }
   }
   else {
      DPrintF("@NewChild:","[Script] Could not gain access to my Owner (%d).", OwnerID);
      return(ERR_ExclusiveDenied);
   }
}

/*****************************************************************************
** Script: NewObject
*/

static ERROR SCRIPT_NewObject(struct Script *Self, APTR Void)
{
   Self->Static = FALSE;

   /*** Assume that the script is in english ***/

   Self->Language[0] = 'e';
   Self->Language[1] = 'n';
   Self->Language[2] = 'g';
   Self->Language[3] = 0;

   StrCopy("lang", Self->LanguageDir, sizeof(Self->LanguageDir));

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  OpeningTag
Short: Activates a Script.
###END###

*****************************************************************************/

static ERROR SCRIPT_OpeningTag(struct Script *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Script: ReleaseObject()
*/

static ERROR SCRIPT_ReleaseObject(struct Script *Self, APTR Void)
{
   if (Self->LocalArgs)  { ReleaseMemory(Self->LocalArgs);  Self->LocalArgs = NULL; }
   if (Self->Location)   { ReleaseMemory(Self->Location);   Self->Location = NULL; }
   if (Self->ObjectList) { ReleaseMemory(Self->ObjectList); Self->ObjectList = NULL; }
   if (Self->Procedure)  { ReleaseMemory(Self->Procedure);  Self->Procedure = NULL; }
   if (Self->Service)    { ReleaseMemory(Self->Service);    Self->Service = NULL; }
   if (Self->String)     { ReleaseMemory(Self->String);     Self->String = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: Script arguments can be set through this action.
###END###

*****************************************************************************/

#define MAX_ARGS 40

static ERROR SCRIPT_SetUnlistedField(struct Script *Self, struct acSetUnlistedField *Args)
{
   STRING str, Arg, field, value;
   MEMORYID ArgMID;
   WORD i, valuelen;

   /* Please note that it is okay to set zero-length string values
   ** (this has its uses in some scripts).
   */

   if ((!Args->Field) OR (!Args->Value)) return(ERR_Args);
   if (!Args->Field[0]) return(ERR_Args);

   field = Args->Field;
   value = Args->Value;

   if (!Self->LocalArgsMID) {
      if (AllocMemory(sizeof(MEMORYID) * MAX_ARGS, Self->Head.MemFlags, (APTR *)&Self->LocalArgs, &Self->LocalArgsMID) != ERR_Okay) {
         return(ERR_Memory);
      }
      Self->TotalArgs = 0;
   }
   else if (Self->TotalArgs >= MAX_ARGS) {
      DPrintF("@SetField:","[Script:%d] Maximum number of arguments reached (%d) for \"%s\".", Self->Head.UniqueID, MAX_ARGS, Self->Location);
      return(ERR_Failed);
   }

   /* Check if the argument refers back to itself (e.g. 'location' = '{location}' would cause a loop-back)
   ** Set the field value to a null-string if such an occurance is detected.
   */

   if (value[0] IS '{') {
      for (i=0; (field[i]) AND (value[i+1]); i++) {
         if (LCase(field[i]) != LCase(value[i+1])) break;
      }
      if ((!field[i]) AND (value[i+1] IS '}') AND (!value[i+2])) {
         DPrintF("@SetUnlistedField:","[Script] Warning: Resetting looped argument '%s = %s'", field, value);
         value = "";
      }
   }

   /* Calculate the length of the argument value, giving consideration to
   ** backslashes because we'll double them up.
   */

   valuelen = 0;
   for (i=0; value[i]; i++) if (value[i] IS '\\') valuelen += 2; else valuelen++;

   if (AllocMemory(StrLength(field) + valuelen + 4,
                   Self->Head.MemFlags|MEM_STRING, (APTR *)&Arg, &ArgMID) IS ERR_Okay) {
      str = Arg;
      *str++ = '{';
      for (i=0; field[i]; i++) *str++ = field[i];
      *str++ = '}';
      *str++ = 0;

      for (i=0; value[i]; i++) {
         if (value[i] IS '\\') { *str++ = '\\'; *str++ = '\\'; }
         else *str++ = value[i];
      }
      *str = 0;

      /*** Check if the argument already exists and overwrite it if it does ***/

      for (i=0; i < Self->TotalArgs; i++) {
         if ((str = AccessArg(Self, i))) {
            if (StrCompare(Arg, str, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
               ReleaseArg(Self, str);

               if (Self->Head.MemFlags & MEM_PUBLIC) {
                  FreeMemoryID(Self->LocalArgs[i]);
                  Self->LocalArgs[i] = ArgMID;
               }
               else {
                  FreeMemory((APTR)Self->LocalArgs[i]);
                  Self->LocalArgs[i] = (LONG)Arg;
               }
               break;
            }
            ReleaseArg(Self, str);
         }
      }

      /*** Insert the argument as a new entry if an existing match wasn't found ***/

      if (i IS Self->TotalArgs) {
         if (Self->Head.MemFlags & MEM_PUBLIC) Self->LocalArgs[Self->TotalArgs] = ArgMID;
         else Self->LocalArgs[Self->TotalArgs] = (LONG)Arg;
         Self->TotalArgs++;
      }

      ReleaseMemory(Arg);
   }
   else return(ERR_Memory);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Cache
Short:  Enables caching for static scripts.
Type:   BOOLEAN
Status: Get/Set

###DESCRIPTION###
<p>By default, static scripts will reload their original script file each
time that they are activated (assuming that the source file has changed).
You can turn off this behaviour by setting the Cache option to TRUE.  In
cache mode, the script will load the source file once and will never check
for changes to the original data source.</p>

<p>Caching does not apply in circumstances where the script source has
originated from a string-based statement.</p>

###END###

*****************************************************************************/

static ERROR GET_Cache(struct Script *Self, LONG *Value)
{
   if (Self->Flags & SCF_CACHE) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Cache(struct Script *Self, LONG *Value)
{
   if (*Value IS TRUE) Self->Flags |= SCF_CACHE;
   else Self->Flags &= ~SCF_CACHE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags.
Prefix: SCF
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Optional flags supported by the Script class can be specified here.
Currently available flags are listed in the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">OBJECTLIST&nbsp;&nbsp;</td><td>If you would like an object list to be generated when a Script object is executed, you need to set this flag.  By default object lists are not created during the activation process as they can slow the routine slightly.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###PRIVATE###
Name:   LocalArgs
Short:  A storage area for Script arguments.
Type:   MEMORYID
Status: Read

###DESCRIPTION###
<p>This field is reserved for use by sub-classes of the Script class.  It
stores arguments that have been set through the unlisted field mechanism.</p>

###END###

*****************************************************************************/

static ERROR GET_LocalArgs(struct Script *Self, MEMORYID **Value)
{
   if (Self->LocalArgs) {
      *Value = Self->LocalArgs;
      return(ERR_Okay);
   }
   else if (!Self->LocalArgsMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->LocalArgsMID, MEM_READWRITE, (void **)&Self->LocalArgs) IS ERR_Okay) {
      *Value = Self->LocalArgs;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

/*****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    The location of the file that is to be processed as a script.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>The location of an object script file must be specified in this field prior
to initialisation.  If the field is not set then the initialisation process
will fail.</p>

<p>Special parameters can also be passed to the script when setting the
location.  The name of an executable procedure may be passed by following the
location with a semicolon, then the name of the procedure to execute.  Arguments
can also be passed to the script by following this with a second semicolon, then
a sequence of arguments, each separated with a comma.  The following string
illustrates the format used:</p>

<pre>
  dir:location;procedure;arg1=val1,arg2,arg3=val2
</pre>

<p>A target for the script may be specified by using the 'target' argument in
the parameter list (value must refer to a valid existing object).</p>

<p>Furthermore, the STRING: assignment is useful if you wish to pass a
raw DML statement directly through this field.  The following example illustrates
valid formatting: "STRING:<dml>...script...</dml>".</p>

###SEE ALSO###
Field: Procedure

###END###

*****************************************************************************/

static ERROR GET_Location(struct Script *Self, STRING *Value)
{
   if (Self->Location) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else if (!Self->LocationMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->LocationMID, MEM_READ, (void **)&Self->Location) IS ERR_Okay) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Location(struct Script *Self, STRING Value)
{
   struct acSetUnlistedField Args;
   WORD i, j, len;

   if (Self->LocationMID) {
      /* If the location has already been set, throw the
      ** value to SetUnlistedField instead.
      */
      if ((Value) AND (*Value)) {
         Args.Field = "Location";
         Args.Value = Value;
         return(Action(AC_SetUnlistedField, Self, &Args));
      }
   }
   else {
      if (Self->Location)    { ReleaseMemory(Self->Location);   Self->Location = NULL; }
      if (Self->LocationMID) { FreeMemoryID(Self->LocationMID); Self->LocationMID = NULL; }

      if ((Value) AND (*Value)) {
         if (StrCompare("STRING:", Value, 7, NULL) IS ERR_Okay) {
            return(SET_String(Self, Value + 7));
         }

         for (len=0; (Value[len]) AND (Value[len] != ';'); len++);
         if (AllocMemory(len+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Location, &Self->LocationMID) IS ERR_Okay) {
            for (i=0; i < len; i++) Self->Location[i] = Value[i];
            Self->Location[i] = 0;

            /* If a semi-colon has been used, this indicates that a procedure
            ** follows the filename.
            */

            if (Value[i] IS ';') {
               UBYTE buffer[800], arg[100], argval[400];

               i++;
               while ((Value[i]) AND (Value[i] <= 0x20)) i++;
               for (j=0; (Value[i]) AND (Value[i] > 0x20) AND (Value[i] != ';'); j++) buffer[j] = Value[i++];
               buffer[j] = 0;
               if (buffer[0]) SET_Procedure(Self, buffer);

               /*** The presence of an opening bracket precedes a series of arguments ***/

               if (Value[i] IS ';') {
                  i++;

                  while (Value[i]) {
                     while ((Value[i]) AND (Value[i] <= 0x20)) i++;
                     while (Value[i] IS ',') {
                        i++;
                        while ((Value[i]) AND (Value[i] <= 0x20)) i++;
                     }

                     /*** Extract arg name ***/

                     for (j=0; (Value[i] != ',') AND (Value[i] != '=') AND (Value[i] > 0x20); j++) arg[j] = Value[i++];
                     arg[j] = 0;

                     while ((Value[i]) AND (Value[i] <= 0x20)) i++;

                     /*** Extract arg value ***/

                     argval[0] = '1';
                     argval[1] = 0;
                     if (Value[i] IS '=') {
                        i++;
                        while ((Value[i]) AND (Value[i] <= 0x20)) i++;
                        if (Value[i] IS '"') {
                           i++;
                           for (j=0; (Value[i]) AND (Value[i] != '"'); j++) argval[j] = Value[i++];
                           argval[j] = 0;
                        }
                        else {
                           for (j=0; (Value[i]) AND (Value[i] != ','); j++) argval[j] = Value[i++];
                           argval[j] = 0;
                        }
                     }

                     if (StrCompare("target", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        SetField(Self, FID_Target, FT_STRING, argval);
                     }
                     else SetUnlistedField(Self, arg, argval);
                  }
               }
            }
         }
         else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Field: Name
*/

static ERROR SET_Name(struct Script *Self, STRING Name)
{
   struct acSetUnlistedField Args;

   if (Name) {
      SetName(Self, Name);

      Args.Field = "Name";
      Args.Value = Name;
      return(SCRIPT_SetUnlistedField(Self, &Args));
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ObjectList
Short:  Lists all objects that have been created during the activation of a Script.
Type:   OBJECTID *
Status: Read

###DESCRIPTION###
<p>If you have set the OBJECTLIST <a href="#Flags">Flag</a> then you can read
this field to retrieve a complete list of objects that were created during the
activation of a Script.  The list is a straight array of OBJECTID values and is
terminated with a NULL value to indicate an end to the list.  The OBJECTID's
are listed in the original order of object creation.</p>

###END###

*****************************************************************************/

static ERROR GET_ObjectList(struct Script *Self, OBJECTID **Value)
{
   if (Self->ObjectList) {
      *Value = Self->ObjectList;
      return(ERR_Okay);
   }
   else if (!Self->ObjectListMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ObjectListMID, MEM_READ, (void **)&Self->ObjectList) IS ERR_Okay) {
      *Value = Self->ObjectList;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

/*****************************************************************************

###FIELD###
Name:   PrimaryObject
Short:  Identifies the primary object of a Script.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>Some object scripts will define a 'primary object' - an object which is
considered to be the most important out of all the objects created by a script
file.  Many graphical scripts, such as &quot;templates:window.dml&quot;,
&quot;templates:button.dml&quot; and &quot;templates:input.dml&quot; will do
this as a development aid.  In DML, primary objects are defined with an
asterisk just before a new object definition.  For example:</p>

<pre>
   &lt;*render width="140" height="60/&gt;
</pre>

<p>The exact purpose of a primary object is for the developer of the script file
to decide.  In most cases the presence of a primary object will save you from
having to trawl through an <a href="#ObjectList">ObjectList</a> to find a
specific object created by a Script.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Procedure
Short:  Allows you to specify a procedure to be executed from within a script.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Sometimes scripts are split into several procedures or functions that can
be executed independently from the 'main' area of the script.  If a script
that you have loaded contains procedures, you can set the Procedure field
to execute a specific routine whenever the script is activated with the
Activate action.</p>

<p>If this field is not set, the first procedure in the script, or the 'main'
procedure (as defined by the script type) is executed by default.</p>

###END###

*****************************************************************************/

static ERROR GET_Procedure(struct Script *Self, STRING *Value)
{
   if (Self->Procedure) {
      *Value = Self->Procedure;
      return(ERR_Okay);
   }
   else if (!Self->ProcedureMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ProcedureMID, MEM_READ, (void **)&Self->Procedure) IS ERR_Okay) {
      *Value = Self->Procedure;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
   return(ERR_Okay);
}

static ERROR SET_Procedure(struct Script *Self, STRING Value)
{
   LONG i;

   if (Self->Procedure)    { ReleaseMemory(Self->Procedure);   Self->Procedure = NULL; }
   if (Self->ProcedureMID) { FreeMemoryID(Self->ProcedureMID); Self->ProcedureMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Procedure, &Self->ProcedureMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Procedure[i] = Value[i];
         Self->Procedure[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###PRIVATE###
Name:   RegisterObject
Short:  Adds an ObjectID to the Script's ObjectList array.
Type:   OBJECTID
Status: Set

###DESCRIPTION###
<p>The RegisterObject field is used during script processing to add newly
allocated objects to the <a href="#ObjectList">ObjectList</a> array.
It is intended for use only by sub-classes of the Script class.</p>

###END###

*****************************************************************************/

static ERROR SET_RegisterObject(struct Script *Self, OBJECTID *ObjectID)
{
   OBJECTID *newlist;
   MEMORYID listmid;
   LONG i;

   if (!(Self->Flags & SCF_OBJECTLIST)) return(ERR_Okay);

   if (!Self->ObjectListMID) {
      if (AllocMemory(sizeof(OBJECTID)*2, Self->Head.MemFlags|MEM_NOCLEAR, (APTR *)&Self->ObjectList, &Self->ObjectListMID) IS ERR_Okay) {
         Self->ObjectList[0] = *ObjectID;
         Self->ObjectList[1] = NULL;
         Self->TotalObjects = 1;
         ReleaseMemory(Self->ObjectList);
      }
   }
   else {
      if (!Self->ObjectList) {
         if (AccessMemory(Self->ObjectListMID, MEM_READ, (void **)&Self->ObjectList) != ERR_Okay) {
            return(ERR_AccessMemory);
         }
      }

      if (AllocMemory(sizeof(OBJECTID) * (Self->TotalObjects + 2), Self->Head.MemFlags|MEM_NOCLEAR,
                      (APTR *)&newlist, &listmid) IS ERR_Okay) {
         for (i=0; i < Self->TotalObjects; i++) newlist[i] = Self->ObjectList[i];
         newlist[Self->TotalObjects++] = *ObjectID;
         newlist[Self->TotalObjects]   = NULL;

         ReleaseMemory(Self->ObjectList);
         FreeMemoryID(Self->ObjectListMID);

         Self->ObjectList    = newlist;
         Self->ObjectListMID = listmid;
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Repeat
Short:  This field can be used to activate a Script multiple times.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you want to execute an object script more than once, the quickest way
to do it is to set the Repeat field.  Simply set the field to the number of
times that you would like to have the script file processed before calling
the Activate action.</p>

###SEE ALSO###
Field: RepeatStart

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   RepeatStart
Short:  Determines the index at which to start a Script's repetition process.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you have set the <a href="#Repeat">Repeat</a> field to execute an
object script multiple times, the repetition count will start at zero and
increment until it reaches the value specified in the Repeat field.  If
you want to start the count from a value other than zero, you can set this
field to that index value.  For instance, if you set the Repeat field to a
value of 8 and the RepeatStart field to a value of 2, the Script will
be processed 6 times.</p>

<p>Setting this field has a direct impact on the Index reference that can be
used in object scripts (refer to the DML Whitepaper for information on
repetition and indexes).</p>

###SEE ALSO###
Field: Repeat

###END###

*****************************************************************************/

/*****************************************************************************

###PRIVATE###
Name:   Service
Short:  ...
Type:   STRING
Status: Get/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_Service(struct Script *Self, STRING *Value)
{
   if (Self->Service) {
      *Value = Self->Service;
      return(ERR_Okay);
   }
   else if (!Self->ServiceMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ServiceMID, MEM_READ, (void **)&Self->Service) IS ERR_Okay) {
      *Value = Self->Service;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_Service(struct Script *Self, STRING Value)
{
   LONG i;

   if (Self->Service)    { ReleaseMemory(Self->Service);   Self->Service = NULL; }
   if (Self->ServiceMID) { FreeMemoryID(Self->ServiceMID); Self->ServiceMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Service, &Self->ServiceMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Service[i] = Value[i];
         Self->Service[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make a Script object static.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>By default, a Script object will execute itself and then self-destruct when
a closing tag is received.  If you would rather that the object stays in the
system, set this field to TRUE.  If you do this, the only way to get the
Script object to process is to call the Activate action.</p>

###END###

*****************************************************************************/

static ERROR SET_Static(struct Script *Self, LONG *Value)
{
   Self->Static = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   String
Short:  May be set to a string to translate if the script data is not in a file.
Type:   STRING
Status: Get/Set

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_String(struct Script *Self, STRING *Value)
{
   if (Self->String) {
      *Value = Self->String;
      return(ERR_Okay);
   }
   else if (!Self->StringMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->StringMID, MEM_READ, (void **)&Self->String) IS ERR_Okay) {
      *Value = Self->String;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError(Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_String(struct Script *Self, STRING Value)
{
   LONG i;

   if (Self->String)    { ReleaseMemory(Self->String);   Self->String = NULL; }
   if (Self->StringMID) { FreeMemoryID(Self->StringMID); Self->StringMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->String, &Self->StringMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->String[i] = Value[i];
         Self->String[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Target
Short:  Defines the container that the Script objects will be initialised to.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>This field should be set to refer to the container that you want the
Script objects to be initialised to.  If you do not set this field then the
root-level objects in the Script will be initialised to the Script's
personal container.</p>

###END###

*****************************************************************************/

static ERROR SET_Target(struct Script *Self, OBJECTID *Value)
{
   Self->TargetID = *Value;
   DPrintF("Set.Target:","[Script] Target: %d", *Value);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   TotalArgs
Short:  Reflects the total number of arguments used on a Script object.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>The total number of arguments that have been set on a Script object through
the unlisted field mechanism are reflected in the value of this field.  If you
have not set any arguments then the field value will be zero.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   TotalObjects
Short:  Reflects the total number of objects created during a Script's activation process.
Type:   LONG
Status: Read

###DESCRIPTION###
<p>If you set the OBJECTLIST <a href="#Flags">Flag</a> prior to a Script
object's activation, this field will reflect the total number of objects
created during the activation process.  If you did not set the OBJECTLIST
flag then this field will be set to zero.</p>

###END###

*****************************************************************************/

/*****************************************************************************

                       "Burning inside, I feel the call
                        Blood of the Ancient Ones now boils
                        At war - War with the race
                        Father rise and the world be one"

                             - Morbid Angel

*****************************************************************************/

#include "functions.c"
