/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Fade
Version:   1.0
Status:    Stable
Date:      May 2003
ID:        ID_FADE
Category:  Effect
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2003.  All rights reserved.
Short:     Adds fading effects to rendered objects.
Keywords:  fade, in, out, translucent, cross-fade, red, green, blue, fade, colour

###DESCRIPTION###
<p>The Fade class provides fading effects that you can apply to rendered objects.
You can use it for two purposes; 1. For fading layers into the display; 2. For
fading layers out of the display.  This class is very simple in its construct
and is intended to provide a convenience to script developers that want to add
fade effects to their work.</p>

<p>The following example illustrates how to use this class to fade a new
drawable into the display:</p>

<pre>
  &lt;render x="30" y="50" width="100" height="200"&gt;
    &lt;fade type="fadein"/&gt;
  &lt;/render&gt;
</pre>

<p>The fade object will set the opacity of the render object to zero on
initialisation, then it will wait for the render object to show itself.  Once
this occurs, the fade is enabled and the opacity will increase until it
reaches its maximum.  At that point, the fade object is no longer needed and
will automatically self-destruct.</p>

<p>You can also place objects inside a fade object to be executed on completion
of the fade process (keep in mind that you must set them as 'static' and remember
that they will be destroyed immediately after being activated).</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/render.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_FADE 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FADE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003.  All rights reserved.",
   "May 2003",
   "Fade"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR FadeClass = 0;
static LONG glBreakMessageID = 0;

/*****************************************************************************
** Fade class definition.
*/

enum {
   FAL_FADEIN=1,
   FAL_FADEOUT
};

struct Fade {
   OBJECT_HEADER
   OBJECTID DrawableID;
   LONG     Flags;
   FLOAT    Seconds;
   LONG     Static;
   LONG     Type;
   FLOAT    Delay;

   PRIVATE_FIELDS
   UBYTE Fade;
   FLOAT Counter;
};

#define FAF_WAIT     0x00000001
#define FAF_ACTIVATE 0x00000002  /* Activate children on completion of the fade */

static struct FieldDef FadeFlags[] = {
   { "WAIT", FAF_WAIT },
   { "ACTIVATE", FAF_ACTIVATE },
   { NULL, NULL }
};

static struct FieldDef FadeType[] = {
   { "FADEIN",  FAL_FADEIN },
   { "FADEOUT", FAL_FADEOUT },
   { NULL, NULL }
};

static struct FieldArray FadeFields[] = {
   { "Drawable", 0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&FadeFlags, NULL, NULL },
   { "Seconds",  0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   { "Static",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Type",     0, FDF_LONG|FDF_LOOKUP|FDF_RI, (LONG)&FadeType, NULL, NULL },
   { "Delay",    0, FDF_FLOAT|FDF_RW,     0, NULL, NULL },
   END_FIELD
};

static ERROR FADE_Activate(struct Fade *, APTR);
static ERROR FADE_ClosingTag(struct Fade *, APTR);
static ERROR FADE_Deactivate(struct Fade *, APTR);
static ERROR FADE_Free(struct Fade *, APTR);
static ERROR FADE_Init(struct Fade *, APTR);
static ERROR FADE_NewObject(struct Fade *, APTR);
static ERROR FADE_Timer(struct Fade *, struct acTimer *);

static struct ActionArray FadeActions[] = {
   { AC_Activate,   FADE_Activate },
   { AC_ClosingTag, FADE_ClosingTag },
   { AC_Deactivate, FADE_Deactivate },
   { AC_Free,       FADE_Free },
   { AC_Init,       FADE_Init },
   { AC_NewObject,  FADE_NewObject },
   { AC_Timer,      FADE_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   glBreakMessageID = AllocateID(IDTYPE_MESSAGE);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FadeClass, NULL,
      FID_BaseClassID|TLONG, ID_FADE,
      FID_Version|TFLOAT,    VER_FADE,
      FID_Name|TSTRING,      "Fade",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      FadeActions,
      FID_Fields|TPTR,       FadeFields,
      FID_Size|TLONG,        sizeof(struct Fade),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (FadeClass) { Action(AC_Free, FadeClass, NULL); FadeClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Fade: Activate
*/

static ERROR FADE_Activate(struct Fade *Self, APTR Void)
{
   struct Message message;
   struct NextMsg nextmsg;

   if (Self->Fade IS FAL_FADEIN) DPrintF("~Activate()","[Fade:%d] Fade-In", Self->Head.UniqueID);
   else if (Self->Fade IS FAL_FADEOUT) DPrintF("~Activate()","[Fade:%d] Fade-Out", Self->Head.UniqueID);
   else DPrintF("~Activate()","[Fade:%d]", Self->Head.UniqueID);

   /*** Initialise the Timer ***/

   SubscribeTimer(Self->Head.UniqueID, 20);

   if (Self->Flags & FAF_WAIT) {
      /*** Wait for the fade to execute before returning ***/
      while (ProcessMessages(&nextmsg, MF_WAIT) IS ERR_Okay) {
         if (nextmsg.Count > 0) {
            if (GetMessage(NULL, NULL, NULL, &message, sizeof(struct Message)) IS ERR_Okay) {
               if ((message.Type IS MSGID_QUIT) OR (message.Type IS glBreakMessageID)) break;
               else DPrintF("@Activate:","[Fade] Received unknown message type #%d.", message.Type);
            }
         }
      }
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Fade: ClosingTag
*/

static ERROR FADE_ClosingTag(struct Fade *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Fade: Deactivate
*/

static ERROR FADE_Deactivate(struct Fade *Self, APTR Void)
{
   LONG count;
   WORD i;

   /*** Activate children on completion of the fade process ***/
   
   if (Self->Flags & FAF_ACTIVATE) {
      if ((count = TotalChildren(Self->Head.UniqueID))) {
         struct ChildEntry list[count];
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
      }
   }

   UnsubscribeTimer(Self->Head.UniqueID);
   acFree(Self);
   return(ERR_Okay);
}

/*****************************************************************************
** Fade: Free
*/

static ERROR FADE_Free(struct Fade *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);
   return(ERR_Okay);
}

/*****************************************************************************
** Fade: Init
*/

static ERROR FADE_Init(struct Fade *Self, APTR Void)
{
   struct mtSetRenderOpacity setopacity;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   if (Self->Type IS FAL_FADEIN) {
      setopacity.Value      = 0;
      setopacity.Adjustment = 0;
      if (ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &setopacity) != ERR_Okay) {
         DPrintF("@Init:","[Fade] Render object does not allow opacity adjustments.");
         return(ERR_Failed);
      }
   }
   else if (Self->Type IS FAL_FADEOUT) {

   }
   else return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));

   return(ERR_Okay);
}

/*****************************************************************************
** Fade: NewObject
*/

static ERROR FADE_NewObject(struct Fade *Self, APTR Void)
{
   Self->Seconds = 1.0;
   return(ERR_Okay);
}

/*****************************************************************************
** Action: Timer
*/

static ERROR FADE_Timer(struct Fade *Self, struct acTimer *Args)
{
   struct mtSetRenderOpacity setopacity;

   if (!Args) return(ObjectError(Self, ERH_Timer, ERR_Args));

   if (Self->Delay > 0.0) {
      if (Self->Delay * 100 > Args->TotalElapsed) {
         return(ERR_Okay);
      }
      else {
         Self->Delay = 0;
         SubscribeTimer(Self->Head.UniqueID, 20); /* Resubscribe to the timer to clear the TotalElapsed counter */
         return(ERR_Okay);
      }
   }

   if (Self->Type IS FAL_FADEIN) {
      setopacity.Value = (Args->TotalElapsed * 100) / (Self->Seconds * 1000);
      if (setopacity.Value >= 100) setopacity.Value = 100;

      setopacity.Adjustment = 0;
      ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &setopacity);

      if (setopacity.Value >= 100) Action(AC_Deactivate, Self, NULL);
   }
   else if (Self->Type IS FAL_FADEOUT) {
      setopacity.Value = 100 - ((Args->TotalElapsed * 100) / (Self->Seconds * 1000));
      if (setopacity.Value < 1) setopacity.Value = 0;

      setopacity.Adjustment = 0;
      ActionMsg(MT_SetRenderOpacity, Self->DrawableID, &setopacity);

      if (setopacity.Value < 1) Action(AC_Deactivate, Self, NULL);
   }
   else UnsubscribeTimer(Self->Head.UniqueID);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Delay
Short:  The start of the fade effect can be delayed by setting this field.
Type:   FLOAT
Status: Read/Init

###DESCRIPTION###
<p>If you would like to delay the start of the fade effect, you can set this
field to the total number of seconds to pause for before the fade starts.
The value is stored in floating point format to allow you to delay for
sub-second intervals.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Sets the drawable targetted by the fade effect.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a fade object it should be contained by a
drawable area, such as a Render object.  However there may be situations 
when you want to initialise a fade object to a non-graphical container.  If 
this is the case then you must set the Drawable field so that the fade object
knows which drawable it should be targetting.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags are set here.
Prefix: FAF
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Special flags supported by the Fade class are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">WAIT</td><td>Tells the object to complete the entirety of the fade before normal processing can resume.</td></tr>
<tr><td width="10%">ACTIVATE</td><td>If set, the fade object will activate all of its children on completion of the fade.</td></tr>
</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Seconds
Short:  The number of seconds over which the fade must complete its processing.
Type:   FLOAT
Status: Read/Set

###DESCRIPTION###
<p>By default, a fade will start and finish within the space of one second.
You can adjust the speed of the fade by changing the number of seconds allowed
for it to execute.  The value is accepted in floating-point format so that you can attain
a high level of accuracy in your timing.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make a Fade object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>To make a fade object static, set this field to TRUE.  If the Static
field is not set, then the object will self-destruct when the fade process
is completed.  Static objects require that you manually start the fade
process with an Activate call.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Type
Short:  The type of fade (fade-in, fade-out) is defined here.
Prefix: FAL
Type:   LONG/LOOKUP
Status: Read/Init

###DESCRIPTION###
<p>Before initialisation of a fade object, you will need to determine whether
or not the target drawable will be faded-in or faded-out of the display.
The only valid settings for this field are FADEIN and FADEOUT.</p>

###END###

*****************************************************************************/
